import xgboost as xgb
from sklearn.metrics import mean_squared_error
import numpy as np

def train_xgboost_model(X_train, y_train, X_test, y_test):
    """
    Train an XGBoost regression model.

    Args:
    - X_train (pd.DataFrame): Features of the training set.
    - y_train (pd.Series): Target variable of the training set.
    - X_test (pd.DataFrame): Features of the test set.
    - y_test (pd.Series): Target variable of the test set.

    Returns:
    - xgb.XGBRegressor: Trained XGBoost model.
    """
    reg = xgb.XGBRegressor(base_score=0.5, booster='gbtree',
                           n_estimators=1000,
                           early_stopping_rounds=50,
                           objective='reg:linear',
                           max_depth=3,
                           learning_rate=0.01)
    reg.fit(X_train, y_train,
            eval_set=[(X_train, y_train), (X_test, y_test)],
            verbose=100)
    return reg
    

def make_predictions(model, X_test):
    """
    Make predictions using a trained XGBoost model.

    Args:
    - model (xgb.XGBRegressor): Trained XGBoost model.
    - X_test (pd.DataFrame): Features of the test set.

    Returns:
    - np.array: Predicted values.
    """
    return model.predict(X_test)

def calculate_rmse(y_true, y_pred):
    """
    Calculate the root mean squared error (RMSE).

    Args:
    - y_true (pd.Series): True target values.
    - y_pred (np.array): Predicted values.

    Returns:
    - float: RMSE value.
    """
    return np.sqrt(mean_squared_error(y_true, y_pred))





import pandas as pd
import numpy as np
import datetime

def create_ordenes_financieras(df, month_to_predict, ordenes_aprobadas=None, service=None):
    """
    Create financial orders based on predicted values and average daily weights.

    Args:
    - df (pd.DataFrame): Dataframe containing order information.
    - month_to_predict (pd.DataFrame): Dataframe containing predicted values for month.
    - ordenes_aprobadas (int, optional): Number of approved orders. If provided, the orders will be distributed based on this value. If not, it will distribute the orders based on the predicted values.


    Returns:
    - pd.DataFrame: DataFrame with financial/forecasted orders.
    """
    if service == "SS":
        orders = df.copy()
        orders = orders[orders["COUNTRY"] != "BR"]
        start_date = datetime.datetime.strptime("2024-01-01", "%Y-%m-%d").date()
        orders = orders[orders["FECHA"] >= start_date]

        orders['FECHA'] = pd.to_datetime(orders['FECHA'])
        pivot_table = orders.pivot_table(index='FECHA', columns='COUNTRY', values='ORDERS', aggfunc='mean', fill_value=0)
        SS = pivot_table.sum(axis=1)
        SS.name = 'SS'
        pivot_table_with_total = pd.merge(pivot_table, SS.to_frame(), on="FECHA", how="inner")
        pivot_table_percentage = pivot_table_with_total.div(pivot_table_with_total['SS'], axis=0) * 100
        pivot_table_percentage = pivot_table_percentage.round(2)
        average_weight_by_day = pivot_table_percentage.groupby(pivot_table_percentage.index.day).mean()
        # average_weight_by_day = average_weight_by_day[:-1] # quitar el 31 de month
        date_strings = [f'2024-07-{day}' for day in average_weight_by_day.index]
        average_weight_by_day.index = pd.to_datetime(date_strings, format='%Y-%m-%d')
        average_weight_by_day.index = average_weight_by_day.index.strftime('%Y-%m-%d')

        # series_reindexed = month_to_predict

        series_reindexed = month_to_predict.reindex(average_weight_by_day.index, fill_value=0)
        # if there are diferent numbers of approved orders then use it.
        if ordenes_aprobadas:
            ordenes_aprobadas = ordenes_aprobadas
            prop_ordenes_dias_mes = series_reindexed/series_reindexed.sum()
            series_reindexed = prop_ordenes_dias_mes*ordenes_aprobadas
        # series_reindexed_modif.head()
        series_reindexed.index = pd.to_datetime(series_reindexed.index, format='%Y-%m-%d')
        average_weight_by_day.set_index(pd.date_range("2024-07-01","2024-07-31"), inplace=True)
        ordenes_financieras = average_weight_by_day.mul(series_reindexed, axis=0) / 100
        orden_dias = ['CO', 'MX', 'EC', 'PE', 'CL', 'AR', 'UY', 'CR', 'SS']
        ordenes_financieras = ordenes_financieras.filter(orden_dias)
        ordenes_financieras = ordenes_financieras.reset_index()
        ordenes_financieras.rename(columns={"index": "FECHA"}, inplace=True)
        
    elif service == "BR":
        orders = df.copy()
        orders = orders[orders["COUNTRY"] == "BR"]
        start_date = datetime.datetime.strptime("2024-01-01", "%Y-%m-%d").date()
        orders = orders[orders["FECHA"] >= start_date]

        orders['FECHA'] = pd.to_datetime(orders['FECHA'])
        pivot_table = orders.pivot_table(index='FECHA', columns='COUNTRY', values='ORDERS', aggfunc='mean', fill_value=0)
        SS = pivot_table.sum(axis=1)
        SS.name = 'SS'
        pivot_table_with_total = pd.merge(pivot_table, SS.to_frame(), on="FECHA", how="inner")
        pivot_table_percentage = pivot_table_with_total.div(pivot_table_with_total['SS'], axis=0) * 100
        pivot_table_percentage = pivot_table_percentage.round(2)
        average_weight_by_day = pivot_table_percentage.groupby(pivot_table_percentage.index.day).mean()
        average_weight_by_day = average_weight_by_day[:-1] # quitar el 31 de month
        date_strings = [f'2024-04-{day}' for day in average_weight_by_day.index]
        average_weight_by_day.index = pd.to_datetime(date_strings, format='%Y-%m-%d')
        average_weight_by_day.index = average_weight_by_day.index.strftime('%Y-%m-%d')

        df_month_to_predict = month_to_predict['prediction']
        series_reindexed = df_month_to_predict.reindex(average_weight_by_day.index, fill_value=0)
        # if there are diferent numbers of approved orders then use it.
        if ordenes_aprobadas:
            ordenes_aprobadas = ordenes_aprobadas
            prop_ordenes_dias_mes = series_reindexed/series_reindexed.sum()
            series_reindexed = prop_ordenes_dias_mes*ordenes_aprobadas
        # series_reindexed_modif.head()
        ordenes_financieras = average_weight_by_day.mul(series_reindexed, axis=0) / 100
        orden_dias = ['CO', 'MX', 'EC', 'PE', 'CL', 'AR', 'UY', 'CR', 'SS']
        ordenes_financieras = ordenes_financieras.filter(orden_dias)
        ordenes_financieras = ordenes_financieras.reset_index()
        ordenes_financieras.rename(columns={"index": "FECHA"}, inplace=True)
    
    return ordenes_financieras

def create_curva_ordenes(df):
    """
    Create order curve based on order distribution throughout the week.

    Args:
    - df (pd.DataFrame): Dataframe containing order information.

    Returns:
    - pd.DataFrame: DataFrame with order curve.
    """

    orders = df.copy()
    orders = orders[orders["COUNTRY"] != "BR"]
    start_date = datetime.datetime.strptime("2024-01-01", "%Y-%m-%d").date()
    orders = orders[orders["FECHA"] >= start_date]

    orders['FECHA'] = pd.to_datetime(orders['FECHA'])
    orders['INTERVALO'] = pd.to_datetime(orders['INTERVALO'], format='%H:%M:%S').dt.time
    orders['DAY_OF_WEEK'] = orders['FECHA'].dt.day_name()

    pivot_table = orders.pivot_table(index='INTERVALO', columns='DAY_OF_WEEK', values='ORDERS', aggfunc='mean', fill_value=0)
    weekdays = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']
    pivot_table.columns = pd.CategoricalIndex(pivot_table.columns, categories=weekdays, ordered=True)
    pivot_table = pivot_table.sort_index(axis=1)

    total_general = pivot_table.sum(axis=0)
    total_general.name = 'Total general'

    pivot_table_with_total = pd.concat([pivot_table, total_general.to_frame().T])

    pivot_table_percentage = pivot_table_with_total.div(pivot_table_with_total.loc['Total general']) * 100
    pivot_table_percentage = pivot_table_percentage.round(2)
    curva_ordenes = pivot_table_percentage.drop(pivot_table_percentage.tail(1).index)  # drop last row
    curva_ordenes = curva_ordenes.reset_index()
    curva_ordenes.rename(columns={"index": "intervalo"}, inplace=True)
    curva_ordenes.set_index("intervalo", inplace=True)
    curva_ordenes = curva_ordenes / 100
    
    return curva_ordenes


def distribute_orders(ordenes_financieras, curva_ordenes):
    """
    Distribute financial orders according to the order curve.

    Args:
    - ordenes_financieras (pd.DataFrame): Dataframe containing financial orders.
    - curva_ordenes (pd.DataFrame): Dataframe containing order curve.

    Returns:
    - pd.DataFrame: DataFrame with distributed orders by day and interval.
    """

    index = pd.MultiIndex.from_product([ordenes_financieras['FECHA'].unique(), curva_ordenes.index], names=['FECHA', 'Intervalo'])
    df3 = pd.DataFrame(index=index, columns=ordenes_financieras.columns[1:], data=np.zeros((len(index), len(ordenes_financieras.columns[1:]))))
    df3 = df3.reset_index()

    def get_day_of_week(date):
        return pd.to_datetime(date).strftime('%A')

    for i, row in ordenes_financieras.iterrows():
        day_of_week = get_day_of_week(row['FECHA'])
        values = row[1:].values
        for j, interval in enumerate(curva_ordenes.index):
            proportion = curva_ordenes.loc[interval, day_of_week]
            df3.loc[(df3['FECHA'] == row['FECHA']) & (df3['Intervalo'] == interval), ordenes_financieras.columns[1:]] += values * proportion
    
    return df3



from data_processing import *
from modeling import *
from export import *
from order_distribution import distribute_orders
from order_distribution import create_ordenes_financieras, create_curva_ordenes

# Path to the Parquet file containing order data
# file_path = "C:/Users/jonathan.marin/Documents/GitHub/rappi/wfm/libreria/one_year_and_a_half.parquet.gzip"
date_range = ['2024-01-01', '2024-04-30']
service = "SS"
# Path to the CSV file containing special dates
#file_path_sp = "C:/Users/jonathan.marin/Documents/GitHub/rappi/wfm/Calendario Rappi - BD_Feriados.csv"

# Start date for the dataset
startdate = "2024-01-01"

# Read and sort order data
df_ordered = read_and_sort_orders(date_range)

# Preprocess special dates
special_days = preprocess_special_dates(date_range)

# Filter special dates from the ordered dataframe
df_filtered = filter_special_dates(df_ordered, special_days)

# Pivot the filtered dataframe to get orders by region and day
orders = pivot_orders(df_filtered, service=service)

# Set the index of orders dataframe to "FECHA" column
data = orders.set_index("FECHA")

# Split the data into training and testing sets
train, test = split_train_test_data(data, startdate)

# Create time series features for training and testing sets
train = create_features(train)
test = create_features(test)

# Features and target variable
FEATURES = ['dayofweek', 'quarter', 'month', 'year', 'dayofyear', 'dayofmonth', 'weekofyear']
TARGET = service

# Extract features and target variable for training and testing sets
X_train = train[FEATURES]
y_train = train[TARGET]

X_test = test[FEATURES]
y_test = test[TARGET]

# Train an XGBoost model
model = train_xgboost_model(X_train, y_train, X_test, y_test)

# Make predictions using the trained model
predictions = make_predictions(model, X_test)

# Add predictions to the test dataframe
test['prediction'] = predictions

# Calculate RMSE score
score = calculate_rmse(y_test, predictions)

# Select data for next month
month_to_predict = test[test.index >= '2024-04-01']

# Create financial orders based on predicted values and average daily weights
ordenes_financieras = create_ordenes_financieras(df_ordered, month_to_predict,service=service, ordenes_aprobadas=14000000)

# Create an order curve based on order distribution throughout the week
curva_ordenes = create_curva_ordenes(df_ordered)

# Distribute financial orders according to the order curve
df3 = distribute_orders(ordenes_financieras, curva_ordenes)
print(df3.head())
print("Done!")



import pandas as pd
import numpy as np


def orders_dist(df, region=None, country=None, aggfunc='sum', dist=False, month=False):
    # Convert 'FECHA' column to datetime
    df['FECHA'] = pd.to_datetime(df['FECHA'])

    # Convert region and country to list if they are not already
    if isinstance(region, str):
        region = [region]
    if isinstance(country, str):
        country = [country]

    # Filter DataFrame if region and/or country are provided
    if region:
        df = df[df['REGION'].isin(region)]
    if country:
        df = df[df['COUNTRY'].isin(country)]

    # Determine the index for the pivot table based on the month parameter
    if month:
        # Extract unique month names from the 'FECHA' column
        unique_months = df['FECHA'].dt.strftime('%B').unique()
        index = df['FECHA'].dt.strftime('%B')
    else:
        index = df['FECHA'].dt.strftime('%A')

    # Pivot DataFrame and calculate distribution of orders by day of the week or month name
    pivot_table = df.pivot_table(
        index=index,
        columns=['REGION', 'COUNTRY'],
        values='ORDERS',
        aggfunc=aggfunc,
        fill_value=np.nan
    )

    # Reorder the index based on the unique months present in the data
    if month:
        # Convert the index to an ordered categorical type for correct month sorting
        months = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December']
        pivot_table.index = pd.CategoricalIndex(pivot_table.index, categories=months, ordered=True)
        pivot_table = pivot_table.sort_index()
    else:
        # Convert the index to an ordered categorical type for correct weekday sorting
        weekdays = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']
        pivot_table.index = pd.CategoricalIndex(pivot_table.index, categories=weekdays, ordered=True)
        pivot_table = pivot_table.sort_index()

    # Compute row totals (grand totals for each day or month)
    pivot_table['Total'] = pivot_table.sum(axis=1)

    # Compute column totals (grand totals for each region and country)
    pivot_table.loc['Grand Total'] = pivot_table.sum()
    
    # Calculate percentage distribution if dist is True
    if dist:
        for col in pivot_table.columns[:-1]: # Exclude 'Total' and 'Grand Total' columns
            pivot_table[col] = (pivot_table[col] / pivot_table['Total']) * 100
        
        # Calculate the percentage for each day or month
        pivot_table['Total'] = (pivot_table['Total'] / pivot_table['Total'].iloc[-1]) * 100

    pivot_table.replace(0, np.nan, inplace=True)

    return pivot_table


import snowflake.connector
import os
import json


def get_project_settings(import_filepath="settings.json"):
    """
    Function to import settings from settings.json
    :param import_filepath: path to settings.json
    :return: settings as a dictionary object
    """
    if os.path.exists(import_filepath):
        try:
            with open(import_filepath, "r") as f:
                project_settings = json.load(f)
            return project_settings
        except Exception as e:
            raise ImportError(f"Error reading settings.json: {e}")
    else:
        raise ImportError("settings.json does not exist at provided location")


def snowflake_connection(project_settings=get_project_settings(), AUTO_COMMIT=True):
    conn = snowflake.connector.connect(
        user=project_settings["snowflake"]["user"],
        password=project_settings["snowflake"]["password"],
        account=project_settings["snowflake"]["account"],
        authenticator=project_settings["snowflake"]["authenticator"],
        database=project_settings["snowflake"]["database"],
        autocommit=True if AUTO_COMMIT == True else False
    )
    return conn


# Function to disconnect the database connection and clear cache
def disconnect_and_clear_cache(conn):
    cur = conn.cursor()
    cur.close()
    # Disconnect the database connection
    conn.close()



import pandas as pd
from dbconn import *
from fetch_data import *
from wfm_functions import *

import pandas as pd
from dbconn import snowflake_connection
from fetch_data import fetch_orders_data

def read_and_sort_orders(date_range=None, file_path=None):
    """
    Read and sort orders data either from a database using a specified date range or from a parquet file.

    Args:
    - date_range (str, optional): A string representing the date range in "YYYY-MM-DD,YYYY-MM-DD" format.
    - file_path (str, optional): Path to the parquet file containing orders data.

    Returns:
    - pd.DataFrame: A DataFrame containing the sorted orders data.

    Raises:
    - ValueError: If neither date_range nor file_path is provided.
    """
    if date_range is not None:
        # If date_range is provided, fetch data from the database
        conn = snowflake_connection()
        df = fetch_orders_data(conn, date_range=date_range, gmv=True, interval=True)
    elif file_path is not None:
        # If file_path is provided, read data from the parquet file
        df = pd.read_parquet(file_path)
    else:
        # If neither date_range nor file_path is provided, raise an error
        raise ValueError("Either date_range or file_path must be provided")

    # Sort the DataFrame by date, interval, and country
    df_ordered = df.sort_values(by=["FECHA", "INTERVALO", "COUNTRY"])
    return df_ordered

import pandas as pd

def preprocess_special_dates(date_range=None, file_path_sp=None):
    """
    Read a CSV file containing special dates and preprocess the dates.

    Args:
    - file_path (str): Path to the file.

    Returns:
    - list: A list of special dates.
    """
    if date_range is not None:
        # If date_range is provided, fetch data from the database
        conn = snowflake_connection()
        special = fetch_special_days(conn,date_range=date_range)
        special.rename(columns={'DATE':'Fecha','EVENT':'FERIADO','COUNTRY':'PAIS','TIER_OPS':'Criticidad'}, inplace=True)
        special=special.filter(['PAIS', 'FERIADO', 'Fecha', 'Criticidad'])

    elif file_path_sp is not None:
        # If file_path_sp is provided, read data from the parquet file
        special = pd.read_csv(file_path_sp)
        # Sort the DataFrame by date, interval, and country
        special['Fecha'] = pd.to_datetime(special['Fecha'], format='%d/%m/%Y').dt.date
        special_days = special["Fecha"].to_list()
        # return special_days
    else:
        # If neither date_range nor file_path is provided, raise an error
        raise ValueError("Either date_range or file_path must be provided")
    
    special['Fecha'] = pd.to_datetime(special['Fecha'], format='%d/%m/%Y').dt.date
    special_days = special["Fecha"].to_list()
    return special_days
    
    

def filter_special_dates(df, special_days):
    """
    Filter out special dates from a DataFrame.

    Args:
    - df (pd.DataFrame): Input DataFrame containing orders data.
    - special_days (list): List of special dates to be filtered out.

    Returns:
    - pd.DataFrame: Filtered DataFrame.
    """
    df_filtered = df[~df['FECHA'].isin(special_days)]
    df_filtered.reset_index(drop=True, inplace=True)
    return df_filtered


def pivot_orders(df, service=None):
    """
    Pivot orders data to get total orders per country per date.

    Args:
    - df (pd.DataFrame): Input DataFrame containing orders data.

    Returns:
    - pd.DataFrame: Pivot table with total orders per country per date.
    """
    if service=="SS":
        ordersxregion = pd.pivot_table(df, values='ORDERS', index=['FECHA', 'INTERVALO'], columns=['COUNTRY'], aggfunc='sum')
        ordersxregion = ordersxregion.fillna(0)
        ordersxregion['SS'] = ordersxregion.drop(columns=['BR']).sum(axis=1)
        ordersxregion['BR'] = ordersxregion['BR']
        ordersxregion = ordersxregion.filter(['CO', 'MX', 'EC', 'PE', 'CL', 'AR', 'UY', 'CR', 'BR', 'SS'])
        orders = ordersxregion[["SS", "BR"]].reset_index()
        orders = orders[["FECHA", "SS", "BR"]].groupby(["FECHA"]).sum()
        orders.reset_index(inplace=True)
    
    elif service == "BR":
        ordersxregion = pd.pivot_table(df, values='ORDERS', index=['FECHA', 'INTERVALO'], columns=['COUNTRY'], aggfunc='sum')
        ordersxregion = ordersxregion.fillna(0)
        ordersxregion['SS'] = ordersxregion.drop(columns=['BR']).sum(axis=1)
        ordersxregion['BR'] = ordersxregion['BR']
        ordersxregion = ordersxregion.filter(['BR', 'SS'])
        orders = ordersxregion[["SS", "BR"]].reset_index()
        orders = orders[["FECHA", "SS", "BR"]].groupby(["FECHA"]).sum()
        orders.reset_index(inplace=True)
    else:
        raise ValueError("Either SS or BR must be provided")
    return orders


def create_features(df):
    """
    Create time series features based on the DataFrame index.

    Args:
    - df (pd.DataFrame): Input DataFrame containing orders data.

    Returns:
    - pd.DataFrame: DataFrame with additional time series features.
    """
    df = df.copy()
    df['dayofweek'] = df.index.dayofweek
    df['quarter'] = df.index.quarter
    df['month'] = df.index.month
    df['year'] = df.index.year
    df['dayofyear'] = df.index.dayofyear
    df['dayofmonth'] = df.index.day
    df['weekofyear'] = df.index.isocalendar().week
    return df

def split_train_test_data(data, startdate):
    """
    Split data into train and test sets based on a start date.

    Args:
    - data (pd.DataFrame): Input DataFrame containing orders data.
    - startdate (str): Start date for the test set in "YYYY-MM-DD" format.

    Returns:
    - tuple: A tuple containing the train and test DataFrames.
    """
    data.index = pd.to_datetime(data.index)
    train = data.loc[data.index < startdate]
    test = data.loc[data.index >= startdate]
    return train, test

def create_future_predictions(model, FEATURES, start_for_pred, end_for_pred, frequency):
    idx = pd.date_range(start=start_for_pred, end=end_for_pred, freq=frequency)
    future_data = pd.DataFrame(index=idx)
    future_data = create_features(future_data)
    future_predictions = model.predict(future_data[FEATURES])
    future_data['prediction'] = future_predictions
    month_to_predict = future_data["prediction"]

    return month_to_predict



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    description = f.read()

setup(
    name="wf_rappi",
    version="0.3.2",
    packages=find_packages(),
    install_requires=[
        'numpy>=1.11.1',
        'pandas>=1.0',
        'matplotlib==3.8.3',
        'openpyxl==3.1.2',
        'packaging==23.2',
        'pillow==10.2.0',
        'psycopg2==2.9.9',
        'pyarrow==15.0.0',
        'pydantic==2.5.3',
        'pydantic-core==2.14.6',
        'requests==2.31.0',
        'scikit-learn==1.3.2',
        'snowflake-connector-python==3.7.1',
        'snowflake-snowpark-python==1.13.0',
        'SQLAlchemy==1.4.51',
        'twine==5.1.0',
        'xgboost==2.0.3'
    ],
    entry_points={
        "console_scripts": [
            "wf-rappi = wf_rappi.data_processing:read_and_sort_orders",
        ],
    },
    long_description=description,
    long_description_content_type="text/markdown",
)


from dataclasses import dataclass, field
from typing import List
from functools import lru_cache
import pandas as pd
import os


@dataclass(frozen=True, order=True) # Instances of this class will be immutable and orderable based on their fields
class Orders:
    service_user_types: List[str] = field(
        default_factory=lambda: ['User', 'RT', 'Rest'])
    segment_types: List[str] = field(
        default_factory=lambda: ['Inhouse - Rappi'])


@lru_cache(maxsize=None)  # This decorator enables caching with unlimited size
def fetch_test(conn, service_user_types: tuple = None, segment_types: tuple = None) -> pd.DataFrame:
    # Convert single strings to tuples
    if isinstance(service_user_types, str):
        service_user_types = (service_user_types,)
    if isinstance(segment_types, str):
        segment_types = (segment_types,)

    # Convert lists to tuples to make them hashable
    if service_user_types is not None:
        service_user_types = tuple(service_user_types)
    else:
        service_user_types = tuple(Orders().service_user_types)
    if segment_types is not None:
        segment_types = tuple(segment_types)
    else:
        segment_types = tuple(Orders().segment_types)

    # Dynamically construct the IN clause for service_user_types
    service_user_types_in_clause = ', '.join(['%s'] * len(service_user_types))
    # Dynamically construct the NOT IN clause for segment_types
    segment_types_not_in_clause = ', '.join(['%s'] * len(segment_types))

    # Prepare the SQL query with placeholders for parameters
    query = f"""
    SELECT date_trunc(week, created_at)::date as Date,
           avg(aht_ls/60) AS AHT
    FROM fivetran.support_bi.cs_all_tickets
    WHERE service_user_type IN ({service_user_types_in_clause})
    AND segment_type NOT IN ({segment_types_not_in_clause})
    GROUP BY 1
    """

    # Combine service_user_types and segment_types into a single tuple
    # Convert None to an empty tuple
    params = (service_user_types or ()) + (segment_types or ())

    # Execute the query with the parameters
    with conn.cursor() as cur:
        cur.execute(query, params)
        result = cur.fetch_pandas_all()

    return result


def fetch_orders_data(conn, countries=None, region=None, date_range=None, interval=False, gmv=False) -> pd.DataFrame:
    """
    Function to fetch data from the database
    :param conn: Snowflake connection object
    :param countries: List of countries to filter the data
    :param region: Region to filter the data
    :param date_range: 'YYYY-MM-DD' formatted string to specify the date range, e.g. '2023-01-01', '2023-01-31
    :param interval: Boolean to specify if the data should be grouped by interval
    :param gmv: Boolean to specify if the data should be fetched from the GMV table"""
    if gmv:
        return fetch_orders_gmv(conn, countries, region, date_range, interval)
    else:
        return fetch_orders_ops(conn, countries, region, date_range, interval)


def fetch_orders_gmv(conn, countries=None, region=None, date_range=None, interval=False) -> pd.DataFrame:
    """
    Function to fetch orders GMV from the database
    :param conn: Snowflake connection object
    :param countries: List of countries to filter the data
    :param region: Region to filter the data
    :param date_range: Tuple with start and end dates
    :param interval: Boolean to specify if the data should be grouped by interval
    """
    if interval:
        query = """
        SELECT 
            TIME_SLICE(created_at::TIMESTAMP, 30, 'm')::DATE AS FECHA,
            TIME_SLICE(created_at::TIMESTAMP, 30, 'm')::TIME AS INTERVALO,
            CASE WHEN COUNTRY = 'BR' THEN 'BR' ELSE 'SS' END AS region,
            country,
            COUNT(order_id) AS orders
        FROM 
            fivetran.global_finances.global_orders
        WHERE 
            CREATED_AT::date BETWEEN '2023-01-01' AND CURRENT_DATE()-1
        """
    else:
        query = """
        SELECT 
            created_at::DATE AS FECHA,
            CASE WHEN COUNTRY = 'BR' THEN 'BR' ELSE 'SS' END AS region,
            country,
            COUNT(order_id) AS orders
        FROM 
            fivetran.global_finances.global_orders
        WHERE 
            CREATED_AT::date BETWEEN '2023-01-01' AND CURRENT_DATE()-1
        """

    params = []

    if isinstance(countries, str):
        query += " AND country = %s "
        params.append(countries)
    elif isinstance(countries, list):
        country_placeholders = ', '.join(['%s'] * len(countries))
        query += f" AND country IN ({country_placeholders})"
        params.extend(countries)

    if region is not None:
        query += " AND region = %s "
        params.append(region)

    if date_range is not None:
        start_date, end_date = date_range
        query += " AND created_at::DATE BETWEEN DATE %s AND DATE %s "
        params.extend([start_date, end_date])

    if interval:
        query += "GROUP BY 1, 2, 3, 4"
    else:
        query += "GROUP BY 1, 2, 3"

    # Execute the query with the parameters
    with conn.cursor() as cur:
        cur.execute(query, params)
        result = cur.fetch_pandas_all()

    return result


def fetch_orders_ops(conn, countries=None, region=None, date_range=None, interval=False) -> pd.DataFrame:
    
    """
    Function to fetch orders ops from the database
    :param conn: Snowflake connection object
    :param countries: List of countries to filter the data
    :param region: Region to filter the data
    :param date_range: 'YYYY-MM-DD' formatted string to specify the date range, e.g. '2023-01-01', '2023-01-31
    :param interval: Boolean to specify if the data should be grouped by interval
    
    Returns:
        pd.DataFrame:
            DataFrame containing the fetched data.
    """
    if interval:
        query = """
        SELECT 
            TIME_SLICE(created_at,30,'m')::date FECHA,
            TIME_SLICE(created_at,30,'m')::time INTERVALO,
            CASE WHEN COUNTRY = 'BR' THEN 'BR' ELSE 'SS' END AS region,
            country,
            COUNT(order_id) AS orders
        FROM 
            FIVETRAN.SUPPORT_BI.CS_ORDERS_GENERAL_TBL
        WHERE 
            CREATED_AT::date BETWEEN '2023-01-01' AND CURRENT_DATE()-1
            AND NOT synthetic AND is_ops 
        """
    else:
        query = """
        SELECT 
            created_at::DATE AS FECHA,
            CASE WHEN COUNTRY = 'BR' THEN 'BR' ELSE 'SS' END AS region,
            country,
            COUNT(order_id) AS orders
        FROM 
            FIVETRAN.SUPPORT_BI.CS_ORDERS_GENERAL_TBL
        WHERE 
            CREATED_AT::date BETWEEN '2023-01-01' AND CURRENT_DATE()-1
            AND NOT synthetic AND is_ops 
        """

    params = []

    if isinstance(countries, str):
        query += " AND country = %s "
        params.append(countries)
    elif isinstance(countries, list):
        country_placeholders = ', '.join(['%s'] * len(countries))
        query += f" AND country IN ({country_placeholders})"
        params.extend(countries)

    if region is not None:
        query += " AND region = %s "
        params.append(region)

    if date_range is not None:
        start_date, end_date = date_range
        query += " AND created_at::date BETWEEN DATE %s AND DATE %s "
        params.extend([start_date, end_date])

    query += "GROUP BY FECHA, INTERVALO, region, country" if interval else "GROUP BY FECHA, region, country"

    # Execute the query with the parameters
    with conn.cursor() as cur:
        cur.execute(query, params)
        result = cur.fetch_pandas_all()

    return result

##############
def fetch_special_days(conn, countries=None, region=None, date_range=None, interval=False) -> pd.DataFrame:
    
    """
    Function to fetch specia days from the database
    :param conn: Snowflake connection object
    :param countries: List of countries to filter the data
    :param region: Region to filter the data
    :param date_range: 'YYYY-MM-DD' formatted string to specify the date range, e.g. '2023-01-01', '2023-01-31
    :param interval: Boolean to specify if the data should be grouped by interval
    
    Returns:
        pd.DataFrame:
            DataFrame containing the fetched data.
    """
    if interval:
        query = """
        select
            TIME_SLICE(DATE,30,'m')::date DATE,
            TIME_SLICE(DATE,30,'m')::time INTERVALO,
            event_type as event,
            country,
            CASE WHEN COUNTRY = 'BR' THEN 'BR' ELSE 'SS' END AS region,
            tier_ops,
            tier_forecast,
            growth_orders
        from FIVETRAN.predictions.global_special_days_tier_growths 
        union
        select DATE::date as DATE,
            event_name as event,
            country,
            CASE WHEN COUNTRY = 'BR' THEN 'BR' ELSE 'SS' END AS region,
            tier as tier_ops,
            null as tier_forecast,
            orders_growth_vs_median as growth_orders
        from FIVETRAN.predictions.global_fcst_special_days 
        """
    else:
        query = """
        select DATE::date as DATE,
            event_type as event,
            country,
            CASE WHEN COUNTRY = 'BR' THEN 'BR' ELSE 'SS' END AS region,
            tier_ops,
            tier_forecast,
            growth_orders
        from FIVETRAN.predictions.global_special_days_tier_growths 
        union
        select DATE::date as DATE,
            event_name as event,
            country,
            CASE WHEN COUNTRY = 'BR' THEN 'BR' ELSE 'SS' END AS region,
            tier as tier_ops,
            null as tier_forecast,
            orders_growth_vs_median as growth_orders
        from FIVETRAN.predictions.global_fcst_special_days  
        """

    # params = []

    # if isinstance(countries, str):
    #     query += " AND country = %s "
    #     params.append(countries)
    # elif isinstance(countries, list):
    #     country_placeholders = ', '.join(['%s'] * len(countries))
    #     query += f" AND country IN ({country_placeholders})"
    #     params.extend(countries)

    # if region is not None:
    #     query += " AND region = %s "
    #     params.append(region)

    # if date_range is not None:
    #     start_date, end_date = date_range
    #     query += " AND created_at::date BETWEEN DATE %s AND DATE %s "
    #     params.extend([start_date, end_date])

    # query += "GROUP BY FECHA, INTERVALO, region, country" if interval else "GROUP BY FECHA, region, country"

    # Execute the query with the parameters
    with conn.cursor() as cur:
        cur.execute(query)
        result = cur.fetch_pandas_all()

    return result

import json
def get_project_settings(import_filepath="settings.json"):
    """
    Function to import settings from settings.json
    :param import_filepath: path to settings.json
    :return: settings as a dictionary object
    """
    if os.path.exists(import_filepath):
        try:
            with open(import_filepath, "r") as f:
                content = f.read().strip()
                if not content:
                    raise ImportError(f"Error reading {import_filepath}: File is empty")
                project_settings = json.loads(content)
            return project_settings
        except FileNotFoundError:
            raise ImportError(f"File not found: {import_filepath}")
        except json.JSONDecodeError as e:
            raise ImportError(f"Error decoding JSON from {import_filepath}: {e}")
        except Exception as e:
            raise ImportError(f"Unexpected error reading {import_filepath}: {e}")
    else:
        raise ImportError(f"settings.json does not exist at provided location: {import_filepath}")